'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var mostSubject = require('most-subject');
var $$observable = _interopDefault(require('symbol-observable'));
var most = require('most');
var invariant = _interopDefault(require('invariant'));

//@noflow

const counter = () => {
  let id = 0;
  return () => ++id;
};
function toTag(...tags) {
  return tags.filter(str => str.length > 0).join('/');
}
const nextPayloadID = counter();
const nextEventID = counter();

const PING = '@@effector/ping';
const PONG = '@@effector/pong';
const HALT = '@@effector/halt';
function createHaltAction() {
  return {
    type: HALT,
    payload: undefined
  };
}

function isAction(data) {
  return typeof data === 'object' && data != null && typeof data.type === 'string';
}

function safeDispatch(data, dispatch) {
  if (Array.isArray(data)) {
    data.filter(isAction).forEach(action => dispatch(action));
  } else if (isAction(data)) dispatch(data);

  return data;
}
function port(dispatch, state$, action$) {
  return function epic(handler) {
    const result = handler(action$, state$).multicast();
    result.observe(data => safeDispatch(data, dispatch));
    return result;
  };
}

// import {safeDispatch, port} from './port'

function basicCommon(domainName, name) {
  const type = toTag(domainName, name);
  return {
    eventID: nextEventID(),
    nextSeq: counter(),
    getType: () => type
  };
}
function observable(event, action$) {
  function observableState() {
    return {
      /**
          * The minimal observable subscription method.
          * @param {Object} observer Any object that can be used as an observer.
          * The observer object should have a `next` method.
          * @returns {subscription} An object with an `unsubscribe` method that can
          * be used to unsubscribe the observable from the store, and prevent further
          * emission of values from the observable.
          */
      subscribe(observer) {
        if (typeof observer !== 'object') {
          throw new TypeError('Expected the observer to be an object.');
        }

        return action$.subscribe(observer);
      },

      [O]() {
        return event;
      }

    };
  }

  const O = $$observable;
  event[O] = observableState;
}

function watchWarn(domainName, name, error) {
  console.warn(`


  Watch function should not fail

  Effect:
    ${name}
  Domain:
    ${domainName}
  Error:
    ${error}

`);
}
function implWarn(domainName, name, props) {
  console.warn(`


  Running an effect without implementation

  Name:
    ${name}
  Domain:
    ${domainName}
  Arguments:
    ${props}


`);
}
function implError(domainName, name, props) {
  const message = `Running an effect without implementation

  Name:
    ${name}
  Domain:
    ${domainName}
  Arguments:
    ${props}


`;
  return new Error(message);
}

function createWatcher({
  event,
  domainName,
  name,
  opts
}) {
  return function watch(fn) {
    event.epic((data$, state$) => most.combine((data, state) => ({
      data,
      state
    }), data$, state$).sampleWith(data$).map(({
      data,
      state
    }) => {
      try {
        const result = fn(data, state);
        return result;
      } catch (err) {
        watchFailWarn({
          domainName,
          name,
          mode: opts.watchFailCheck()
        }, err);
      }
    }));
  };
}

function watchFailWarn({
  domainName,
  name,
  mode
}, error) {
  switch (mode) {
    case 'throw':
      {
        throw error;
      }

    case 'warn':
      {
        watchWarn(domainName, name, error);
      }

    case 'off':
    default:
      {
        return;
      }
  }
}

function EventConstructor(domainName, dispatch, getState, state$, events, name, action$ = mostSubject.sync(), config = {
  isPlain: false,
  watchFailCheck: 'warn'
}) {
  const {
    eventID,
    nextSeq,
    getType
  } = basicCommon(domainName, name);
  const handlers = new Set();
  if (!config.isPlain) handlers.add((payload
  /*::, state*/
  ) => action$.next(payload)); // handlers.add((payload, state) => state$.next(state))

  const create = payload => ({
    type: getType(),
    payload,
    meta: {
      index: nextPayloadID(),
      eventID,
      seq: nextSeq(),
      passed: false,
      plain: config.isPlain
    }
  });

  function eventInstance(payload) {
    const toSend = Promise.resolve(payload);
    const result = create(payload);
    let canDispatch = true;
    return { ...result,

      raw() {
        return result;
      },

      send(dispatchHook) {
        if (canDispatch) {
          const dispatcher = dispatchHook ? dispatchHook : dispatch;
          canDispatch = false;
          dispatcher(result);

          for (const handler of handlers) {
            handler(payload, getState());
          } // action$.next(payload)

        }

        return toSend;
      }

    };
  }

  eventInstance.getType = getType; //$off

  eventInstance.toString = getType;
  eventInstance.watch = createWatcher({
    event: eventInstance,
    domainName,
    name,
    opts: {
      watchFailCheck: () => config.watchFailCheck
    }
  });
  eventInstance.epic = port(dispatch, state$, action$);

  eventInstance.subscribe = subscriber => action$.subscribe(subscriber); // eventInstance.port = function port<R>(events$: Stream<R>) {
  //   return events$.observe(data => safeDispatch(data, dispatch))
  // }


  eventInstance.trigger = function trigger(query, eventName = 'trigger') {
    const triggerEvent = EventConstructor(domainName, dispatch, getState, state$, events, [name, eventName].join(' '));
    triggerEvent.watch((_, state) => {
      const queryResult = query(state);
      return eventInstance(queryResult);
    });
    return triggerEvent;
  };

  observable(eventInstance, action$);
  events.set(getType(), eventInstance);
  return eventInstance;
}

function EffectConstructor(domainName, dispatch, getState, state$, events, name, opts, action$ = mostSubject.sync()) {
  const handlers = new Set();
  const {
    eventID,
    nextSeq,
    getType
  } = basicCommon(domainName, name);
  handlers.add((payload, state) => action$.next(payload));
  let using = getWarn({
    domainName,
    name,
    mode: () => opts.unused()
  });

  function use(thunk) {
    using = thunk;
  }

  const create = payload => ({
    type: getType(),
    payload,
    meta: {
      index: nextPayloadID(),
      eventID,
      seq: nextSeq(),
      passed: false
    }
  });

  const effect = params => {
    let run = () => {
      run = () => {};

      const runned = runUsing(params, using);
      runned.then(result => {
        const data = {
          params,
          result
        };
        dispatcher(done(data), dispatch);
        resolve(data);
      }, error => {
        const data = {
          params,
          error
        };
        dispatcher(fail(data), dispatch);
        reject(data);
      });
    };

    const runner = dispatchHook => {
      if (canDispatch) {
        canDispatch = false;
        dispatcher(result, dispatch, dispatchHook);

        for (const handler of handlers) {
          handler(params, getState());
        }

        run(); // action$.next(params)
      }
    };

    let resolve = val => {};

    let reject = val => {};

    const doneP = new Promise(rs => resolve = rs);
    const failP = new Promise(rs => reject = rs);
    const toSend = Promise.resolve(params);
    const result = create(params);
    let canDispatch = true;
    return { ...result,

      raw() {
        return result;
      },

      send(dispatchHook) {
        runner(dispatchHook);
        return toSend;
      },

      done() {
        runner();
        return doneP;
      },

      fail() {
        runner();
        return failP;
      },

      promise() {
        runner();
        return Promise.race([doneP, failP.then(data => Promise.reject(data))]);
      }

    };
  };

  const done = EventConstructor(domainName, dispatch, getState, state$, events, [name, 'done'].join(' '));
  const fail = EventConstructor(domainName, dispatch, getState, state$, events, [name, 'fail'].join(' '));
  effect.use = use;
  effect.getType = getType;
  setToString(effect, getType);
  effect.done = done;
  effect.fail = fail;
  effect.epic = port(dispatch, state$, action$);
  effect.watch = createWatcher({
    event: effect,
    domainName,
    name,
    opts
  });

  effect.trigger = function trigger(query, eventName = 'trigger') {
    const triggerEvent = EventConstructor(domainName, dispatch, getState, state$, events, [name, eventName].join(' '));
    triggerEvent.watch((_, state) => {
      const queryResult = query(state);
      return effect(queryResult);
    });
    return triggerEvent;
  };

  effect.subscribe = subscriber => action$.subscribe(subscriber);

  observable(effect, action$);
  return effect;
}

function setToString(effect, getType) {
  effect.toString = getType;
}

function runUsing(params, using) {
  let response;

  try {
    response = using(params);
  } catch (err) {
    return Promise.reject(err);
  }

  if (typeof response === 'object' && response != null && typeof response.then === 'function') {
    return response;
  }

  return Promise.resolve(response);
}

function dispatcher(value, dispatchDefault, dispatchHook, isBatch = false) {
  if (!isAction$1(value)) {
    return value;
  }

  const dispatch = typeof dispatchHook === 'function' ? dispatchHook : dispatchDefault;
  const result = dispatch(value);
  return result;
}

function isAction$1(data) {
  return typeof data === 'object' && data != null && typeof data.type === 'string';
}

const getWarn = ({
  domainName,
  name,
  mode
}) => function never(props) {
  switch (mode()) {
    case 'throw':
      {
        return Promise.reject(implError(domainName, name, props));
      }

    case 'warn':
      {
        implWarn(domainName, name, props);
      }

    case 'off':
    default:
      {
        return new Promise(() => {});
      }
  }
};

function createDomain(store, domainName = '') {
  const redux = store;
  const config = redux.dispatch({
    type: PING
  }).payload;

  const addPlainHandler = (name, fn) => {
    config.plain.set(name, fn);
  };

  const state$ = most.from(redux).map(() => store.getState()).multicast();
  return DomainConstructor(domainName, redux.dispatch, store.getState, state$, new Map(), addPlainHandler);
}
function createRootDomain(domainName = '') {
  const addPlainHandler = (name, fn) => {
    config.plain.set(name, fn);
  };

  let redux;
  let reduxStore;
  const config = {
    plain: new Map()
  };

  const dispatch = function dispatch(data) {
    return redux.dispatch(data);
  };

  function getState() {
    return reduxStore.getState();
  }

  const state$ = mostSubject.sync();
  let unsubscribe;

  function disablePrevious(previousDispatch) {
    previousDispatch(createHaltAction());

    if (typeof unsubscribe === 'function') {
      unsubscribe();
    }
  }

  function register(store) {
    const previousDispatch = redux != null ? redux.dispatch : v => v;
    redux = reduxStore = store;
    const pong = redux.dispatch({
      type: PING,
      payload: undefined
    });

    if (pong) {
      const payload = pong.payload;
      const plain = payload.plain;
      const oldPlain = config.plain;
      oldPlain.clear(); // for (const [key, val] of oldPlain.entries()) {
      //   // if (!plain.has(key)) {
      //   plain.set(key, val)
      //   // }
      // }

      config.plain = plain;
    }

    disablePrevious(previousDispatch); //$off

    const unsub = most.from(store).map(() => getState()).subscribe({
      next(value) {
        state$.next(value);
      },

      error(err) {},

      complete() {}

    });

    unsubscribe = () => unsub.unsubscribe();
  }

  const domain = DomainConstructor(domainName, dispatch, getState, state$, new Map(), addPlainHandler);
  domain.register = register;
  return domain;
}

function optsReducer([key, value]) {
  if (!(key in this)) return;

  const newVal = () => value;
  this[key] = newVal;
}

function mergeEffectOpts(defaults, opts) {
  if (opts === undefined) return { ...defaults
  };

  if (!(typeof opts === 'object' && opts != null)) {
    throw new Error(`effect options should be object, got ${typeof opts}`);
  }

  const resultObj = { ...defaults
  };
  Object.keys(opts).map(key => [key, opts[key]]).forEach(optsReducer, resultObj);
  return resultObj;
}

function DomainConstructor(domainName, dispatch, getState, state$, events = new Map(), addPlainHandler = () => {}, configGetter) {
  let getConfig;

  if (typeof configGetter !== 'function') {
    const config = {
      unused: () => 'warn',
      watchFailCheck: () => 'warn',
      dispatch
    };

    getConfig = () => config;
  } else getConfig = configGetter;

  const disp = getDispatch(getConfig);
  const effectOpts = {
    unused() {
      return getConfig().unused();
    },

    watchFailCheck() {
      return getConfig().watchFailCheck();
    },

    dispatch: disp
  };

  function recoverable(events$) {
    const str$ = events$.multicast();

    const failureHandler = e => {
      portFailWarn({
        domainName
      }, e);
      return str$;
    };

    return str$.recoverWith(failureHandler);
  }

  return {
    port(events$) {
      return recoverable(events$).observe(data => {
        safeDispatch(data, dispatch);
      });
    },

    register(store) {
      console.warn(`Not implemented`);
    },

    effect(name, opts) {
      const optsFull = mergeEffectOpts(effectOpts, opts);
      return EffectConstructor(domainName, dispatch, getState, state$, events, name, optsFull);
    },

    domain(name) {
      return DomainConstructor(toTag(domainName, name), dispatch, getState, state$, events, addPlainHandler, getConfig);
    },

    event(name) {
      return EventConstructor(domainName, dispatch, getState, state$, events, name);
    },

    typeConstant(name) {
      const action$ = mostSubject.sync();
      addPlainHandler(name, data => {
        if (data.meta == null) data.meta = {};
        data.meta.plain = true;
        const resultEvent = result(data.payload);
        resultEvent.meta.passed = true;
        resultEvent.meta.plain = true;
        action$.next(data.payload);
        getConfig().dispatch(resultEvent);
      });
      const result = EventConstructor('', dispatch, getState, state$, events, name, action$, {
        isPlain: true,
        watchFailCheck: effectOpts.watchFailCheck()
      });
      return result;
    }

  };
}

function getDispatch(getConfig) {
  const dispatch = value => {
    const result = getConfig().dispatch(value);
    return result;
  };

  return dispatch;
}

function portFailWarn({
  domainName
}, error) {
  console.warn(`


  Port stream should not fail.
  Next failure will throw.

  Domain:
    ${domainName}
  Error:
    ${error}

`);
}

function effectorMiddleware() {
  const plainMap = new Map();
  const config = {
    plain: plainMap
  };
  let isActive = true;
  return next => action => {
    if (!isActive) {
      return action;
    }

    if (!(typeof action === 'object' && action != null)) {
      return next(action);
    }

    if (typeof action.type !== 'string') {
      return action;
    }

    if (action.type === PING) {
      return {
        type: PONG,
        payload: config
      };
    }

    if (action.type === HALT) {
      isActive = false;
      plainMap.clear();
      return next(action);
    }

    const handler = plainMap.get(action.type);
    if (!(typeof action.meta === 'object' && action.meta != null)) action.meta = {};
    if (action.meta.passed === true) return action;
    if (handler) action.meta.plain = true;

    if (typeof action.send === 'function') {
      const {
        send,
        meta
      } = action;

      if (handler) {
        meta.passed = true;
        send(next);
        handler(action);
        return action;
      }

      action.send(next);
      return action;
    } else {
      const {
        meta
      } = action;

      if (handler) {
        if (meta.passed !== true) {
          meta.passed = true;
          const result = next(action);
          handler(action);
          return result;
        }

        action.meta.passed = true;
      }

      action.meta.passed = true;
      return next(action);
    }
  };
}
/*
TODO Replace middleware with store enhancer,
as domains needs access to replaceReducers
to handle reducers on their own.
Also, domain might be storeEnhancer itself
*/

function normalizeType(typeOrActionCreator) {
  if (typeOrActionCreator && typeOrActionCreator.getType) {
    return typeOrActionCreator.toString();
  }

  return typeOrActionCreator;
}

function createReducer(defaultState, handlers = {}, thisType) {
  const opts = {
    fallback: null,
    defaultState
  };
  const reducer = Object.assign(reduce, {
    has,
    on,
    off,
    options,
    reset
  });
  const returnType = thisType ? thisType : reducer;

  function has(typeOrActionCreator) {
    return !!handlers[normalizeType(typeOrActionCreator)];
  }

  function reset(typeOrActionCreator) {
    on(typeOrActionCreator, (state, payload, meta) => opts.defaultState);
    return returnType;
  }

  function on(typeOrActionCreator, handler) {
    if (Array.isArray(typeOrActionCreator)) {
      typeOrActionCreator.forEach(action => {
        on(action, handler);
      });
    } else {
      handlers[normalizeType(typeOrActionCreator)] = handler;
    }

    return returnType;
  }

  function off(typeOrActionCreator) {
    if (Array.isArray(typeOrActionCreator)) {
      typeOrActionCreator.forEach(off);
    } else {
      delete handlers[normalizeType(typeOrActionCreator)];
    }

    return returnType;
  }

  function options(newOpts) {
    // $off
    Object.keys(newOpts).forEach(name => opts[name] = newOpts[name]);
    return returnType;
  }

  if (typeof handlers === 'function') {
    const factory = handlers; // $off

    handlers = {};
    factory(on, off);
  }

  function reduce(stateRaw, action) {
    const state = stateRaw === undefined ? opts.defaultState : stateRaw;

    if (!action || typeof action.type !== 'string') {
      return state;
    }

    if (action.type.startsWith('@@redux/')) {
      return state;
    }

    const handler = handlers[String(action.type)] || opts.fallback;

    if (handler) {
      const result = handler(state, action.payload, action.meta);
      if (result === undefined) return opts.defaultState;
      return result;
    }

    return state;
  }

  return reducer;
}

//@noflow
function combine(combine, ...reducers) {
  const values = reducers.map(reducer => reducer(undefined, {
    type: '@@effector/void'
  }));
  const defaultValues = [...values];
  const defaultState = combine(...values);
  let lastState = defaultState;
  const combined = createReducer(defaultState, {}, combinedReducer);

  function combinedReducer(state, action) {
    const changed = new Set();

    for (let i = 0; i < reducers.length; i++) {
      const prevValue = values[i];
      const value = reducers[i](prevValue, action);

      if (value !== prevValue) {
        changed.add(i);
        values[i] = value === undefined ? defaultValues[i] : value;
      }
    }

    const notChanged = changed.size === 0;
    let rawResult;
    if (notChanged) rawResult = lastState;else {
      lastState = rawResult = combine(...values);
    }
    if (rawResult === undefined) lastState = rawResult = defaultState;
    const result = combined(rawResult, action);
    if (result === undefined) return lastState = defaultState;
    return result;
  }

  Object.assign(combinedReducer, combined);
  return combinedReducer;
}

let defaultDomain;

function domainGetter() {
  if (defaultDomain === undefined) {
    const res = createRootDomain();
    defaultDomain = res;
  }

  return defaultDomain;
} // export function


function createEvent$1(name) {
  const domain = domainGetter();
  const event = domain.event(name);
  return event;
}
function createEffect(name) {
  const domain = domainGetter();
  const effect = domain.effect(name);
  return effect;
}

class Collect {
  and(red) {
    return new Collect([...this.parts, red]);
  }

  combine(fn) {
    const reducers = this.parts.length === 0 ? [createReducer()] : this.parts;
    const join = combine;
    const joined = join(fn, ...reducers);
    return joined;
  }

  joint(fn) {
    return this.combine(fn);
  }

  constructor(parts = []) {
    Object.defineProperty(this, "parts", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    this.parts = parts;
  }

}
function collect() {
  return new Collect();
}

class Scope {
  constructor() {
    Object.defineProperty(this, "anyMap", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: new WeakMap()
    });
    Object.defineProperty(this, "dispatchMap", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: new WeakMap()
    });
    Object.defineProperty(this, "eventsMap", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: new WeakMap()
    });
    Object.defineProperty(this, "referenceMap", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: new WeakMap()
    });
  }

}

function genericSafeGet(store, instance, fabric) {
  const result = store.get(instance);
  if (result !== undefined) return result;
  const value = fabric();
  store.set(instance, value);
  return value;
}

function getFromSetMap(setMap, that) {
  return genericSafeGet(setMap, that, () => new Set());
}

function getFromAnyMap(scope, instance) {
  return getFromSetMap(scope.anyMap, instance);
}
function getFromDispatchMap(scope, instance) {
  return getFromSetMap(scope.dispatchMap, instance);
}
function getFromEventsMap(scope, instance) {
  return genericSafeGet(scope.eventsMap, instance, () => new Map());
}
function getListeners(scope, instance, eventName) {
  const events = getFromEventsMap(scope, instance);
  return getFromSetMap(events, eventName);
}
function getReferences(scope, instance) {
  return getFromSetMap(scope.referenceMap, instance);
}

function clearListeners(scope, instance, eventName) {
  if (eventName !== undefined) {
    getListeners(scope, instance, eventName).clear();
  } else {
    getFromAnyMap(scope, instance).clear();

    for (const listeners of getFromEventsMap(scope, instance).values()) {
      listeners.clear();
    }
  }
}
function listenerCount(scope, instance, eventName) {
  let count = getFromAnyMap(scope, instance).size;

  if (eventName !== undefined) {
    return count + getListeners(scope, instance, eventName).size;
  }

  for (const value of getFromEventsMap(scope, instance).values()) {
    count += value.size;
  }

  return count;
}
function onDispatch(scope, instance, fn) {
  getFromDispatchMap(scope, instance).add(fn);
  return () => offDispatch(scope, instance, fn);
}

function offDispatch(scope, instance, fn) {
  getFromDispatchMap(scope, instance).delete(fn);
}

function on(scope, instance, eventName, listener) {
  invariant(instance !== eventName, 'circular reference'); //TODO add support for many listers for one event in references

  getReferences(scope, eventName).add(instance);
  getListeners(scope, instance, eventName).add(listener);
  return () => off(scope, instance, eventName, listener);
}
function off(scope, instance, eventName, listener) {
  getReferences(scope, eventName).delete(instance);
  getListeners(scope, instance, eventName).delete(listener);
}
function once(scope, instance, eventName) {
  return new Promise((rs, rj) => {
    try {
      const off = on(scope, instance, eventName, data => {
        off();
        rs(data);
      });
    } catch (err) {
      rj(err);
    }
  });
}
function onAny(scope, instance, listener) {
  getFromAnyMap(scope, instance).add(listener);
  return () => offAny(scope, instance, listener);
}
function offAny(scope, instance, listener) {
  getFromAnyMap(scope, instance).delete(listener);
}

const resolvedPromise = Promise.resolve();

function emitRefs(scope, eventName, refs, eventData) {
  if (refs.size === 0) return;
  invariant(!refs.has(eventName), 'circular reference');

  for (const instance of [...refs]) {
    emitSync(scope, instance, eventName, eventData);
  }
}

function emitCallbacks(scope, eventName, eventData) {
  const refs = getFromDispatchMap(scope, eventName); //  console.log(eventName, eventData)

  if (refs.size === 0) return;

  for (const instance of [...refs]) {
    instance(eventData, eventName);
  }
}

function dispatchSync(scope, eventName, eventData) {
  const refs = getReferences(scope, eventName);
  emitRefs(scope, eventName, refs, eventData);
  emitCallbacks(scope, eventName, eventData);
}
function emitSync(scope, instance, eventName, eventData) {
  const listeners = [...getListeners(scope, instance, eventName)];
  const anyListeners = [...getFromAnyMap(scope, instance)];

  for (const listener of listeners) {
    listener(eventData, eventName);
  }

  for (const listener of anyListeners) {
    listener(eventData, eventName);
  }

  emitCallbacks(scope, eventName, eventData);
}
async function emit(scope, instance, eventName, eventData) {
  const listeners = getListeners(scope, instance, eventName);
  const anyListeners = getFromAnyMap(scope, instance);
  const staticListeners = [...listeners];
  const staticAnyListeners = [...anyListeners];
  await resolvedPromise;
  emitCallbacks(scope, eventName, eventData);
  return Promise.all([...staticListeners.map(async listener => {
    if (listeners.has(listener)) {
      return listener(eventData, eventName);
    }
  }), ...staticAnyListeners.map(async listener => {
    if (anyListeners.has(listener)) {
      return listener(eventData, eventName);
    }
  })]);
}
async function emitSerial(scope, instance, eventName, eventData) {
  const listeners = getListeners(scope, instance, eventName);
  const anyListeners = getFromAnyMap(scope, instance);
  const staticListeners = [...listeners];
  const staticAnyListeners = [...anyListeners];
  await resolvedPromise;

  for (const listener of staticListeners) {
    if (listeners.has(listener)) {
      await listener(eventData, eventName);
    }
  }

  for (const listener of staticAnyListeners) {
    if (anyListeners.has(listener)) {
      await listener(eventData, eventName);
    }
  }

  emitCallbacks(scope, eventName, eventData);
}

//import invariant from 'invariant'
let id = 0;
const defaultScope = new Scope();
class Emittery {
  constructor() {
    Object.defineProperty(this, "id", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: ++id
    });
  }

  on(eventName, listener, scope = defaultScope) {
    return on(scope, this, eventName, listener);
  }

  off(eventName, listener, scope = defaultScope) {
    off(scope, this, eventName, listener);
  }

  once(eventName, scope = defaultScope) {
    return once(scope, this, eventName);
  }

  emit(eventName, eventData, scope = defaultScope) {
    return emit(scope, this, eventName, eventData);
  }

  emitSerial(eventName, eventData, scope = defaultScope) {
    return emitSerial(scope, this, eventName, eventData);
  }

  emitSync(eventName, eventData, scope = defaultScope) {
    return emitSync(scope, this, eventName, eventData);
  }

  dispatchSync(eventData, scope = defaultScope) {
    return dispatchSync(scope, this, eventData);
  }

  onDispatch(fn, scope = defaultScope) {
    return onDispatch(scope, this, fn);
  }

  onAny(listener, scope = defaultScope) {
    return onAny(scope, this, listener);
  }

  offAny(listener, scope = defaultScope) {
    offAny(scope, this, listener);
  }

  clearListeners(eventName, scope = defaultScope) {
    clearListeners(scope, this, eventName);
  }

  listenerCount(eventName, scope = defaultScope) {
    return listenerCount(scope, this, eventName);
  }

}

function atom(defaultValue) {
  return new Atomic(defaultValue);
}

function remote(baseValue) {
  let value = baseValue;
  return {
    get() {
      return value();
    },

    set(newValue) {
      value = () => newValue;
    }

  };
}

function createAtom(ref) {
  const active = remote(() => true);
  const emission = new Emittery();
  const set = new Emittery();
  const get = new Emittery();
  const offA = emission.on(set, ref.set);
  const offB = emission.on(get, cb => cb(ref.get()));

  const off = () => {
    active.set(false);
    offA();
    offB();
  };

  return {
    get,
    set,
    off,
    active,
    emission
  };
}

function getSafe(cb, get) {
  get.dispatchSync(cb);
}

function getReturn(defaultValue, get) {
  let value = defaultValue;
  getSafe(newValue => {
    value = newValue;
  }, get);
  return value;
}

class Atomic {
  constructor(defaultValue) {
    Object.defineProperty(this, "meta", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    let defValue = defaultValue instanceof Atomic ? () => {
      const result = defaultValue.get();

      defValue = () => result;

      return result;
    } : () => defaultValue;
    const {
      get,
      set,
      off,
      active,
      emission
    } = createAtom(remote(defValue));
    Object.defineProperty(this, 'meta', {
      value() {
        return {
          get,
          set,
          off,
          active,
          emission,
          defaultValue: defValue
        };
      }

    });
  }

  get() {
    const {
      active,
      defaultValue,
      get
    } = this.meta();
    invariant(active.get(), 'inactive atom');
    return getReturn(defaultValue(), get);
  }

  set(newValue) {
    const {
      active,
      set
    } = this.meta();
    invariant(active.get(), 'inactive atom');
    set.dispatchSync(newValue); //  emission.emitSync(set, newValue)
  }

  isActive() {
    return this.meta().active.get();
  }

  off() {
    this.meta().off();
  }

  watch(fn) {
    return this.meta().set.onDispatch(fn);
  }

  map(fn) {
    return new Reference(_ => [fn(_)], this);
  }

  chain(fn) {
    return new Reference(fn, this);
  }

  sample(b, fn) {
    let bVal = b.get();
    const offB = b.watch(b => {
      bVal = b;
    });
    const offA = iterate(this, {
      next(val) {
        const result = fn(val, bVal);
        atom.set(result);
      },

      complete() {
        atom.off();
        offA();
        offB();
      }

    });
    const atom = new Atomic();
    return atom;
  }

  combine(b, fn) {
    //  const halt = () => this.off()
    const set = (a, b) => {
      const result = fn(a, b);
      atom.set(result);
    };

    const offB = b.watch(b => {
      set(this.get(), b);
    });
    const offA = iterate(this, {
      next(val) {
        set(val, b.get());
      },

      complete() {
        atom.off();
        offA();
        offB();
      }

    });
    const atom = new Atomic();
    return atom;
  }

  update(fn) {
    const oldValue = this.get();
    const newValue = fn(oldValue);
    if (oldValue === newValue) return this;
    if (newValue === undefined) return this;
    this.set(newValue);
    return this;
  }

}

function iterate(ref, observable) {
  const off = ref.watch(_ => observable.next(_));
  return () => {
    observable.complete();
    off();
  };
}

class Reference extends Atomic {
  constructor(transformer, ref) {
    const defs = [...transformer(ref.get())];
    super(defs[defs.length - 1]);

    const halt = () => this.off();

    const set = t => this.set(t);

    const off = iterate(ref, {
      next(val) {
        for (const result of transformer(val)) {
          set(result);
        }
      },

      complete() {
        halt();
        off();
      }

    });
  }

} // export function from<T>(transformer: Generator<T>, ref: Atomic<T>) {}
 // export function map<A, B>

exports.createDomain = createDomain;
exports.createRootDomain = createRootDomain;
exports.effectorMiddleware = effectorMiddleware;
exports.createReducer = createReducer;
exports.combine = combine;
exports.joint = combine;
exports.createHaltAction = createHaltAction;
exports.createEvent = createEvent$1;
exports.createEffect = createEffect;
exports.mill = collect;
exports.collect = collect;
exports.Atomic = Atomic;
exports.atom = atom;
exports.Reference = Reference;
